/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.email.inboxlistener.engine;


import java.net.URL;
import java.net.MalformedURLException;
import javax.xml.namespace.QName;
import org.apache.axis.AxisFault;
import org.apache.axis.client.Call;
import org.apache.axis.client.Service;



/**
 * This class groups some utility methods for WS invoking and handling
 *
 * @version $Id: WSTools.java,v 1.2 2006/08/08 09:30:32 gibi Exp $
 */
public class WSTools {
    // --------------------------------------------------------------- Constants

    public static final String FAULT_UNAUTHORIZED = "(401)Unauthorized";

    public static final String FAULT_DATA_STORE   = "DataStoreException";

    // ------------------------------------------------------------ Private data

    /**
     * The WS endpoint
     */
    private URL endpoint;

    /**
     * Username for authentication
     */
    private String username;

    /**
     * Password for authentication
     */
    private String password;

    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new WSTools with the given endpoint
     *
     * @param endpoint the endpoint
     */
    public WSTools(String endpoint, String username, String password)
    throws Exception {
        try {
            this.endpoint = new URL(endpoint);
        } catch (MalformedURLException e) {
            throw new Exception("Incorrect admin URL '" + endpoint + "'");
        }
        this.username = username;
        this.password = password;
    }

    /**
     * Invokes the given WS method with the given objects.
     *
     * @param methodName the method to invoke
     * @param args the call arguments
     *
     * @return the object returned by the invoked method
     *
     * @throws AdminExcheption in case of SyncAdmin or administration errors
     */
    public Object invoke(final String methodName, Object[] args)
        throws Exception {
        Service  service = new Service();

        debug("Invoking " + methodName + " on " + endpoint);

        try {
           Call call = (Call) service.createCall();

           mapTypes(call);

           call.setTargetEndpointAddress(endpoint);
           call.setOperationName(methodName);
           call.setUsername(username);
           call.setPassword(password);

           return call.invoke(args);
        } catch (AxisFault f) {

            //
            // The exceptions are handled with the following roles:
            // 1. If there is a cause, we throw a SyncAdmin exception with
            //    CONNECTION_ERROR_MESSAGE. (Usually the cause is a connection exception).
            // 2. If the faultString is FAULT_UNAUTHORIZED we throw an UnauthorizedException
            // 3. If the faultCode is Server.userException we have to check if the exception
            //    is an Exception or a ServerException.
            //    If it is a Exception we get his message from the faultString and
            //    we throw a new Exception with this message.
            //    If it is a ServerException we throw an Exception with
            //    message UNEXPECTED_SERVER_ERROR.
            // 4. Otherwise we throw an Exception with message UNEXPECTED_ERROR.
            //
            String errorMessage = null;


            error(f);


            Throwable cause = f.getCause();
            if (cause != null) {
                debug("Cause: " + f.getCause());
            }

            if (cause != null) {
                throw new Exception("Connection error", cause);
            }

            if (FAULT_UNAUTHORIZED.equals(f.getFaultString())) {

               throw new Exception(username);

           } else if (org.apache.axis.Constants.FAULT_SERVER_USER.equals(f.getFaultCode().getLocalPart())) {

               String faultString  = f.getFaultString();

               if (faultString != null) {
                   if (faultString.startsWith("sync4j.server.admin.Exception:")) {
                       int len = "sync4j.server.admin.Exception: ".length();
                       errorMessage = faultString.substring(len);
                       debug("errorMessage: " + errorMessage);
                   } else {
                       errorMessage = "Unexpected error";
                   }
               }

               throw new Exception(errorMessage);

           } else {
               if (f.getMessage() != null && !f.getMessage().equals("")) {
                   throw new Exception("Unexpected error: " +
                                       f.getMessage(), f);
               } else {
                   throw new Exception("Unexpected error: ", f);
               }
           }

        } catch (Exception e) {
            error(e);

            throw new Exception("Unexpected error: " +
                                e.getMessage(), e);
        }
    }


    // --------------------------------------------------------- Private methods

    /**
     * Maps the java types to the serializer-deserializer to use (the AXIS
     * BeanSerializer/Deserializer).
     *
     * @param call the call on which perform the mappings
     *
     */
    private void mapTypes(Call call) {
    }

    /**
     * Maps a single type.
     *
     * @param call the Call object
     * @param type the class to map
     * @param qname the type QName
     *
     */
    private void mapType(Call call, final Class type, final String qname)  {
        QName q = new QName( "urn:BeanService",  qname);
        call.registerTypeMapping(
            type,
            q,
            new org.apache.axis.encoding.ser.BeanSerializerFactory(type, q),
            new org.apache.axis.encoding.ser.BeanDeserializerFactory(type, q)
        );
    }


    private void debug(String message) {
        System.out.println(message);
    }

    private void error(Throwable t) {
        t.printStackTrace();
    }
}
